@@ -1,5 +1,5 @@
 
-import sys, math
+import sys, os, math
 
 import cpp
 
@@ -16,14 +16,19 @@ CODE = r"""
 #include <CGAL/Polygon_2.h>
 #include <CGAL/Polygon_with_holes_2.h>
 #include <CGAL/Polygon_set_2.h>
+#include <CGAL/Boolean_set_operations_2.h>
+#include <CGAL/minkowski_sum_2.h>
+#include <CGAL/Small_side_angle_bisector_decomposition_2.h>
+#include <CGAL/connect_holes.h>
 
 #include <vector>
 #include <iostream>
 #include <sstream>
 
-typedef CGAL::Exact_predicates_exact_constructions_kernel Kernel;
+//typedef CGAL::Exact_predicates_exact_constructions_kernel Kernel;
 //typedef CGAL::Extended_homogeneous<CGAL::Gmpz>            Kernel;
 //typedef CGAL::Cartesian<CGAL::Gmpq>                       Kernel;
+typedef CGAL::Homogeneous<CGAL::Gmpz>                     Kernel;
 
 typedef Kernel::Point_3                                   Point_3;
 typedef Kernel::Vector_3                                  Vector_3;
@@ -33,10 +38,12 @@ typedef Polyhedron::HalfedgeDS                            HalfedgeDS;
 typedef CGAL::Aff_transformation_3<Kernel>                Aff_transformation_3;
 
 typedef Kernel::Point_2                                   Point_2;
+typedef Kernel::Vector_2                                  Vector_2;
 typedef CGAL::Polygon_2<Kernel>                           Polygon_2;
 typedef CGAL::Polygon_with_holes_2<Kernel>                Polygon_with_holes_2;
 typedef CGAL::Polygon_set_2<Kernel>                       Polygon_set_2;
-
+typedef CGAL::Aff_transformation_2<Kernel>                Aff_transformation_2;
+typedef CGAL::Small_side_angle_bisector_decomposition_2<Kernel> Decomposition_strategy;
 
 class Build_polyhedron : public CGAL::Modifier_base<HalfedgeDS> {
 public:
@@ -81,7 +88,7 @@ include(${CGAL_USE_FILE})
 
 """
 
-M = cpp.Module('__cppcache__', CODE, CMAKE_CODE)
+M = cpp.Module(os.path.join(os.path.expanduser('~'),'.demakein'), CODE, CMAKE_CODE)
 
 def int_round(value):
     return int(math.floor(value+0.5))
@@ -99,7 +106,7 @@ class Shape(object):
         xs = []
         ys = []
         zs = []
-        for tri in self.triangles(): #Possibly slow?
+        for tri in self.iter_triangles(): #Possibly slow?
             for point in tri:
                 xs.append(point[0])
                 ys.append(point[1])
@@ -115,9 +122,16 @@ class Shape(object):
         xmin,xmax,ymin,ymax,zmin,zmax = self.extent()
         return xmax-xmin,ymax-ymin,zmax-zmin
 
-    def triangles(self):
+    def iter_triangles(self):
+        point_cache = { }
+        def get_cache(p):
+            if p not in point_cache:
+                point_cache[p] = p
+            return point_cache[p]
+        
         polyhedron = M.new('Polyhedron()')
-        M('a.interior().convert_to_polyhedron(b)',a=self.nef,b=polyhedron)
+        #M('a.interior().convert_to_polyhedron(b)',a=self.nef,b=polyhedron)
+        M('a.convert_to_polyhedron(b)',a=self.nef,b=polyhedron)
         
         #assert M('(int)a.is_valid(true, 3)',a=polyhedron)
         
@@ -125,24 +139,27 @@ class Shape(object):
         for item in M.iterate('a.facets_begin()','a.facets_end()',a=polyhedron):
             verts = [ ]
             for item2 in M.circulate('a.facet_begin()',a=item):
-                point = M('a.vertex()->point()',a=item2)
-                verts.append((
-                     M('CGAL::to_double(a.x())',a=point),
-                     M('CGAL::to_double(a.y())',a=point),
-                     M('CGAL::to_double(a.z())',a=point),
-                ))
+                #point = M('a.vertex()->point()',a=item2)
+                verts.append(get_cache((
+                     M('CGAL::to_double(a.vertex()->point().x())',a=item2),
+                     M('CGAL::to_double(a.vertex()->point().y())',a=item2),
+                     M('CGAL::to_double(a.vertex()->point().z())',a=item2),
+                )))
                 
             assert len(verts) == 3
-            triangles.append(verts)
-        return triangles
+            yield tuple(verts)
+    
+    def triangles(self):
+        return list(self.iter_triangles())
 
     def save(self, filename):
-        sys.stdout.write(filename)
+        print filename,
         sys.stdout.flush()
-        triangles = self.triangles()
         with open(filename,'wb') as f:
             print >> f, 'solid'
-            for tri in triangles:
+            n = 0
+            for tri in self.iter_triangles():
+                n += 1
                 print >> f, 'facet normal 0 0 0'
                 print >> f, 'outer loop'
                 for vert in tri:
@@ -150,7 +167,7 @@ class Shape(object):
                 print >> f, 'endloop'
                 print >> f, 'endfacet'
             print >> f, 'endsolid'
-        print
+        print n, 'triangles'
 
     def remove(self, other):
         M.do('a -= b',a=self.nef,b=other.nef)
@@ -332,24 +349,49 @@ def show_only(*shapes):
 class Shape_2(object):
     def __init__(self, pset):
         self.pset = pset
+
+    def copy(self):
+        return Shape_2(M.new('Polygon_set_2(a)',a=self.pset))
     
-    def loops(self):
+    def _iter_polygons_with_holes(self):
+        phole_vec = M.new('std::vector<Polygon_with_holes_2>')
+        M.do('a.polygons_with_holes(std::back_inserter(b))', a=self.pset,b=phole_vec)        
+        for phole in M.iterate('a.begin()','a.end()',a=phole_vec):
+            yield phole
+            
+    def loops(self, holes=True):
         def point_decode(point):
             return (
                 M('CGAL::to_double(a.x())',a=point),
                 M('CGAL::to_double(a.y())',a=point)
             )
         def polygon_decode(poly):
-            return [ point_decode(item) for item in M.iterate('a.vertices_begin()','a.vertices_end()',a=poly) ]
+            return shape.Loop([ point_decode(item) for item in M.iterate('a.vertices_begin()','a.vertices_end()',a=poly) ])
     
         loops = [ ]
     
-        phole_vec = M.new('std::vector<Polygon_with_holes_2>')
-        M.do('a.polygons_with_holes(std::back_inserter(b))', a=self.pset,b=phole_vec)        
-        for phole in M.iterate('a.begin()','a.end()',a=phole_vec):
+        for phole in self._iter_polygons_with_holes():
+            assert not M('a.is_unbounded()',a=phole)
             loops.append( polygon_decode( M('a.outer_boundary()',a=phole) ) )
-            loops.extend( polygon_decode(item) for item in  M.iterate('a.holes_begin()','a.holes_end()',a=phole) )
+            if holes:
+                loops.extend( polygon_decode(item) for item in  M.iterate('a.holes_begin()','a.holes_end()',a=phole) )
         return loops
+
+    def loop(self, holes=True):
+        [ result ] = self.loops(holes)
+        return result
+
+    def extent(self):
+        xs = [ ]
+        ys = [ ]
+        for loop in self.loops():
+            for point in loop:
+                xs.append(point[0])
+                ys.append(point[1])
+        return shape.Limits_2(
+            min(xs), max(xs),
+            min(ys), max(ys),
+        )
     
     def to_3(self):
         #TESTME!
@@ -362,9 +404,7 @@ class Shape_2(object):
         
         #result = empty_shape()
         result = None
-        phole_vec = M.new('std::vector<Polygon_with_holes_2>')
-        M.do('a.polygons_with_holes(std::back_inserter(b))', a=self.pset,b=phole_vec)        
-        for phole in M.iterate('a.begin()','a.end()',a=phole_vec):
+        for phole in self._iter_polygons_with_holes():
             phole_3 = polygon_recode( M('a.outer_boundary()',a=phole) )
             for hole in M.iterate('a.holes_begin()','a.holes_end()',a=phole):
                 phole_3.remove( polygon_recode(hole) )
@@ -377,7 +417,6 @@ class Shape_2(object):
             
         return result
         
-
     def remove(self, other):
         M.do('a.difference(b)',a=self.pset,b=other.pset)
 
@@ -392,12 +431,103 @@ class Shape_2(object):
     
     def orientation(self):
         return M('(int)a.orientation()',a=self.pset)
+    
+    def is_empty(self):
+        return M('a.is_empty()',a=self.pset)
+
+
+    def _reduce_things(self, things):
+        if not things:
+            return empty_shape_2()
+        while len(things) > 1:
+            item = things.pop(-1)
+            things[len(things)//2].add(item)
+        return things[0]
 
     def move(self, x,y, accuracy=1<<16):
-        transform = M('Aff_transformation_2(CGAL::TRANSLATION,Vector_2(a,b,c))',
+        #transform = M('Aff_transformation_2(CGAL::TRANSLATION,Vector_2(a,b,c))',
+        #    a=int_round(x*accuracy), b=int_round(y*accuracy), c=accuracy
+        #)
+        offset = M('Vector_2(a,b,c)',
             a=int_round(x*accuracy), b=int_round(y*accuracy), c=accuracy
         )
-        M.do('a.transform(b)',a=self.pset, b=transform)
+        def transform(polygon):
+            vec = M.new('std::vector<Point_2>')
+            for point in M.iterate('a.vertices_begin()','a.vertices_end()',a=polygon):
+                M.do('a.push_back(b+c)',a=vec,b=point,c=offset)
+            return M.new('Polygon_2(a.begin(),a.end())',a=vec)
+        
+        #M.do('a.transform(b)',a=self.pset, b=transform)
+        things = [ ]
+        for phole in self._iter_polygons_with_holes():
+            outside = transform( M('a.outer_boundary()',a=phole) )
+            insides = M.new('std::vector<Polygon_2>')
+            for inside in M.iterate('a.holes_begin()','a.holes_end()',a=phole):
+                inside = transform(inside)
+                M.do('a.push_back(b)',a=insides,b=inside)
+            thing = M.new('Polygon_set_2(Polygon_with_holes_2(a,b.begin(),b.end()))',a=outside,b=insides)
+            things.append(thing)       
+            
+        self.pset = self._reduce_things(things)
+
+    def intersects(self, other):
+        temp = self.copy()
+        temp.clip(other)
+        return not temp.is_empty()
+    
+        #return M('CGAL::do_intersect(a,b)',a=self.pset,b=other.pset)
+        #bholes = list(other._iter_polygons_with_holes())
+        #for ahole in self._iter_polygons_with_holes():
+        #    for bhole in bholes:
+        #        if M('CGAL::do_intersect(a,b)',a=ahole,b=bhole):
+        #            return True
+        #return False
+
+    def minkowski_sum(self, other):
+        things = [ ]
+        my_polys = map(_connect_holes,self._iter_polygons_with_holes())
+        other_polys = map(_connect_holes,other._iter_polygons_with_holes())
+        for a in my_polys:
+            for b in other_polys:
+                things.append(Shape_2(
+                    M.new(
+                        'Polygon_set_2(CGAL::minkowski_sum_2(a,b,Decomposition_strategy()))',
+                        a=a,b=b)
+                    ))
+        return self._reduce_things(things)
+
+    def erosion(self, other):
+        ext1 = self.extent()
+        ext2 = other.extent()
+        #Generous margin
+        r = max(abs(item) for item in list(ext1)+list(ext2))
+        inner_loop = shape.square(r*3)
+        outer_loop = shape.square(r*5)
+        
+        negative = outer_loop.shape_2()
+        negative.remove(self)
+        dilation = negative.minkowski_sum(other)
+        result = inner_loop.shape_2()
+        result.remove(dilation)
+        return result
+    
+    def offset_curve(self, amount, quality=None):
+        """ +ve dilation
+            -ve erosion """
+        if amount == 0.0:
+            return self.copy()
+        elif amount > 0.0:
+            return self.minkowski_sum(shape.circle(amount*2.0, quality).shape_2())
+        else:
+            return self.erosion(shape.circle(amount*-2.0, quality).shape_2())    
+        
+
+def _connect_holes(poly_with_holes):
+    """ Converts a polygon_with_holes to a polygon """
+    vec = M.new('std::vector<Point_2>')
+    M.do('CGAL::connect_holes(a,std::back_inserter(b))',a=poly_with_holes,b=vec)
+    poly = M.new('Polygon_2(a.begin(),a.end())',a=vec)
+    return poly
 
 
 def empty_shape_2():
