@@ -14,7 +14,7 @@ Fingerholes are enumerated from bottom (end-most) to top.
 
 """
 
-import math, os, pickle, sys
+import math, os, pickle, sys, random
 
 import profile, svg
 
@@ -39,6 +39,12 @@ semitone = {
 
 
 def fqc(note):
+    if '*' in note:
+        note, mult = note.split('*')
+        mult = float(mult)
+    else:
+        mult = 1.0
+
     s = semitone[note[0].upper()]
     note = note[1:]
     if note.startswith('b'):
@@ -50,7 +56,7 @@ def fqc(note):
     
     s += 12 * int(note)
     
-    return 440.0 * 2**((s-57)/12.0)
+    return 440.0 * 2**((s-57)/12.0) * mult
 
 
 def wavelength(w, transpose=0):
@@ -313,15 +319,15 @@ def scaler(value):
     @property
     def func(self):
         scale = self.scale
-        return [ item*scale for item in value ]
+        return [ item*scale if item is not None else None for item in value ]
     return func
 
 
 def sqrt_scaler(value):
     @property
     def func(self):
-        scale = self.scale
-        return [ item*math.sqrt(scale) for item in value ]
+        scale = math.sqrt(self.scale)
+        return [ item*scale if item is not None else None for item in value ]
     return func
 
 
@@ -496,60 +502,74 @@ class Instrument_designer(config.Action_with_output_dir):
     
     def constraint_score(self, inst):
         """ Return an amount of constraint dissatisfaction """
-        score = 0.0
+        scores = [ ]
         
-        if inst.length < 0: 
-            score += -inst.length
+        #if inst.length < 0: 
+            #score += -inst.length
+        scores.append(inst.length)
         
         inners = [0.0]+inst.inner_kinks+[inst.length]
         for i in range(len(inners)-1):
             check = inners[i+1] - inners[i] - self.min_inner_fraction_sep[i]*inst.length
-            if check < 0: score += -check 
+            #if check < 0: score += -check 
+            scores.append(check)
         
         outers = [0.0]+inst.outer_kinks+[inst.length]
         for i in range(len(outers)-1):
             check = outers[i+1] - outers[i] - self.min_outer_fraction_sep[i]*inst.length
-            if check < 0: score += -check
+            #if check < 0: score += -check
+            scores.append(check)
         
         if self.n_holes:
             check = inst.hole_positions[0] - self.bottom_clearance_fraction*inst.length
-            if check < 0: score += -check
+            #if check < 0: score += -check
+            scores.append(check)
             
             check = (1.0-self.top_clearance_fraction)*inst.length - inst.hole_positions[-1]
-            if check < 0: score += -check
+            #if check < 0: score += -check
+            scores.append(check)
             
             for i, value in enumerate(self.min_hole_spacing):
                 if value is None: continue
                 check = (inst.hole_positions[i+1]-inst.hole_positions[i]) - value
-                if check < 0: score += -check
+                #if check < 0: score += -check
+                scores.append(check)
             
             for i, value in enumerate(self.max_hole_spacing):
                 if value is None: continue
                 check = value - (inst.hole_positions[i+1]-inst.hole_positions[i])
-                if check < 0: score += -check
+                #if check < 0: score += -check
+                scores.append(check)                
             
             for i, value in enumerate(self.min_hole_diameters):
                 check = inst.hole_diameters[i] - value
-                if check < 0: score += -check
+                #if check < 0: score += -check
+                scores.append(check)
             
             for i, value in enumerate(self.max_hole_diameters):
                 if value is None: continue
                 check = value - inst.hole_diameters[i]
-                if check < 0: score += -check
+                #if check < 0: score += -check
+                scores.append(check)
             
             for i, value in enumerate(self.balance):
                 if value is None: continue
                 check = value * 0.5*(inst.hole_positions[i+2]-inst.hole_positions[i]) - abs(
                     0.5*inst.hole_positions[i]+0.5*inst.hole_positions[i+2]-inst.hole_positions[i+1]
                 )
-                if check < 0: score += -check
+                #if check < 0: score += -check
+                scores.append(check)
 
-        return score
+        negscores = [ -item for item in scores if item < 0.0 ]
+        negscore = sum(negscores) if negscores else 0.0
+        #posscores = [ (item+1e-30)**-1 for item in scores if item >= 0.0 ]
+        #posscore = 1.0/sum(posscores) if posscores else 1e30
+        return negscore
 
     
     def score(self, inst):
         score = 0.0
-        scale = 0.0
+        div = 0.0
         
         inst.prepare()
         
@@ -557,29 +577,60 @@ class Instrument_designer(config.Action_with_output_dir):
         for note, fingers in self.fingerings:
             w1 = wavelength(note, self.transpose)
             w2, grad = inst.true_wavelength_near(w1, fingers, self.max_grad)
-            #score += (abs(w1-w2)/w1) ** 2
-            #scale += 1 #w1 ** 2
-            #score += ( math.log(w1)-math.log(w2) )**2
             diff = abs(math.log(w1)-math.log(w2))*s
-            score += diff*diff*diff / ((100.0 + diff)**2)
+            #weight = w1
+            weight = 1.0
+            score += weight * diff**3 / (1.0 + (diff/20.0)**2)
+            div += weight
             
-        return (score/len(self.fingerings))
+        return (score/div)**(1.0/3)
         
         #return ( (score/scale) ** (1.0/2) )*100.0
 
+    def _constrainer(self, state_vec):
+        return self.constraint_score(self.unpack(state_vec))
 
-    def _opt_score(self, state_vec):
-        inst = self.unpack(state_vec)
-        
-        cs = self.constraint_score(inst)
-        if cs:
-            result = (cs, 0.0)        
-        else:
-            result = (0.0, self.score(inst))                
-        return result
+    def _scorer(self, state_vec):
+        return self.score(self.unpack(state_vec))
+
+    #def _opt_score(self, state_vec):
+    #    inst = self.unpack(state_vec)
+    #    
+    #    cs = self.constraint_score(inst)
+    #    if cs:
+    #        result = (cs, 0.0)
+    #    else:
+    #        result = (0.0, self.score(inst))
+    #    return result
 
     
-    def _save(self, state_vec):
+    def _draw(self, diagram, state_vec, color='#000000', red_color='#ff0000'):
+        instrument = self.unpack( state_vec )
+        instrument.prepare()
+        
+        for i in range(self.n_holes):
+            diagram.circle(0.0, -instrument.inner_hole_positions[i], instrument.hole_diameters[i],
+                           red_color)
+            diagram.circle(0.0, -instrument.hole_positions[i], instrument.hole_diameters[i], color)
+        diagram.profile(instrument.outer, color)
+        diagram.profile(instrument.stepped_inner, color)
+        
+        if self.closed_top:
+            d = instrument.stepped_inner(instrument.length)
+            diagram.line([(-0.5*d,-instrument.length),(0.5*d,-instrument.length)], color)
+        
+        tick_x = instrument.outer.maximum() * -0.625
+        for pos in instrument.inner_kinks:
+            diagram.line([(tick_x,-pos),(tick_x-5,-pos)], color)
+        for pos in instrument.inner.pos[1:-1]:
+            diagram.line([(tick_x-2,-pos),(tick_x-3,-pos)], color)
+        for pos in instrument.outer_kinks:
+            diagram.line([(tick_x-10,-pos),(tick_x-15,-pos)], color)
+        for pos in instrument.outer.pos[1:-1]:
+            diagram.line([(tick_x-12,-pos),(tick_x-13,-pos)], color)
+
+    
+    def _save(self, state_vec, other_vecs=[]):
         self.instrument = self.unpack( state_vec )
         
         with open(os.path.join(self.output_dir, 'data.pickle'), 'wb') as f:
@@ -602,16 +653,22 @@ class Instrument_designer(config.Action_with_output_dir):
             mod_instrument.prepare()
         
         diagram = svg.SVG()
-        for i in range(self.n_holes):
-            diagram.circle(0.0, -self.instrument.inner_hole_positions[i], self.instrument.hole_diameters[i],
-                           '#ff0000')
-            diagram.circle(0.0, -self.instrument.hole_positions[i], self.instrument.hole_diameters[i])
-        diagram.profile(self.instrument.outer)
-        diagram.profile(self.instrument.stepped_inner)
         
-        if self.closed_top:
-            d = self.instrument.stepped_inner(self.instrument.length)
-            diagram.line([(-0.5*d,-self.instrument.length),(0.5*d,-self.instrument.length)])
+        for vec in random.sample(other_vecs, min(20,len(other_vecs))):
+            self._draw(diagram, vec, '#aaaaaa', '#ffaaaa')
+        
+        self._draw(diagram, state_vec)
+        
+        #for i in range(self.n_holes):
+        #    diagram.circle(0.0, -self.instrument.inner_hole_positions[i], self.instrument.hole_diameters[i],
+        #                   '#ff0000')
+        #    diagram.circle(0.0, -self.instrument.hole_positions[i], self.instrument.hole_diameters[i])
+        #diagram.profile(self.instrument.outer)
+        #diagram.profile(self.instrument.stepped_inner)
+        #
+        #if self.closed_top:
+        #    d = self.instrument.stepped_inner(self.instrument.length)
+        #    diagram.line([(-0.5*d,-self.instrument.length),(0.5*d,-self.instrument.length)])
             
         text_x = diagram.max_x * 1.25
         text_y = 0
@@ -668,16 +725,6 @@ class Instrument_designer(config.Action_with_output_dir):
             
             
         
-        tick_x = diagram.min_x * 1.25
-        for pos in self.instrument.inner_kinks:
-            diagram.line([(tick_x,-pos),(tick_x-5,-pos)])
-        for pos in self.instrument.inner.pos[1:-1]:
-            diagram.line([(tick_x-2,-pos),(tick_x-3,-pos)])
-        for pos in self.instrument.outer_kinks:
-            diagram.line([(tick_x-10,-pos),(tick_x-15,-pos)])
-        for pos in self.instrument.outer.pos[1:-1]:
-            diagram.line([(tick_x-12,-pos),(tick_x-13,-pos)])
-        
         diagram.text(0.0, 20.0, 'Outer diameters: %s' % ', '.join('%.1fmm'%item for item in self.outer_diameters))
         diagram.text(0.0, 40.0, 'Inner diameters: %s' % ', '.join('%.1fmm'%item for item in self.inner_diameters))
         
@@ -698,7 +745,7 @@ class Instrument_designer(config.Action_with_output_dir):
             os.mkdir(self.output_dir)
 
         state_vec = self.initial_state_vec
-        state_vec = optimize.improve('Optimize', self._opt_score, state_vec, monitor=self._save)
+        state_vec = optimize.improve(self.shell_name(), self._constrainer, self._scorer, state_vec, monitor=self._save)
         
         self._save(state_vec)
         
